# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2024-05-20 15:21:57 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.2"
githash = "355a87d6e1b2a3cfc6ed5bf2da76bd6b8249533d"


major = 3
minor = 0
bugfix = 2

release = True
timestamp = datetime.datetime(2024, 5, 20, 15, 21, 57)
debug = False

astropy_helpers_version = ""

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
